# Copyright IBM All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0
import asyncio
import logging
from hfc.fabric.peer import create_peer
from hfc.fabric.transaction.tx_context import create_tx_context
from hfc.fabric.transaction.tx_proposal_request import create_tx_prop_req, \
    CC_TYPE_GOLANG, CC_INSTANTIATE, CC_INSTALL, TXProposalRequest
from hfc.util.crypto.crypto import ecies
from hfc.util.utils import send_transaction, build_tx_req
from test.integration.utils import get_peer_org_user, \
    BaseTestCase
from test.integration.config import E2E_CONFIG
from test.integration.e2e_utils import build_channel_request, \
    build_join_channel_req, get_stream_result

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
test_network = E2E_CONFIG['test-network']
CC_PATH = 'github.com/example_cc'
CC_NAME = 'example_cc'
CC_VERSION = '1.0'


class ChaincodeInstantiateTest(BaseTestCase):

    def test_instantiate_chaincode(self):
        loop = asyncio.get_event_loop()

        channel = self.client.new_channel(self.channel_name)
        org1 = 'org1.example.com'
        peer_config = test_network['org1.example.com']['peers']['peer0']
        tls_cacerts = peer_config['tls_cacerts']
        opts = (('grpc.ssl_target_name_override',
                 peer_config['server_hostname']),)
        endpoint = peer_config['grpc_request_endpoint']

        peer = create_peer(endpoint=endpoint,
                           tls_cacerts=tls_cacerts,
                           opts=opts)

        org1_admin = get_peer_org_user(org1, 'Admin',
                                       self.client.state_store)
        crypto = ecies()

        # for chain code install
        tran_prop_req_in = create_tx_prop_req(
            prop_type=CC_INSTALL,
            cc_path=CC_PATH,
            cc_type=CC_TYPE_GOLANG,
            cc_name=CC_NAME,
            cc_version=CC_VERSION
        )

        # install chain code
        tx_context_in = create_tx_context(org1_admin,
                                          crypto,
                                          tran_prop_req_in)

        # for chain code deploy
        args = ['a', '100', 'b', '40']
        tran_prop_req_dep = create_tx_prop_req(
            prop_type=CC_INSTANTIATE,
            cc_type=CC_TYPE_GOLANG,
            cc_name=CC_NAME,
            cc_version=CC_VERSION,
            args=args,
            fcn='init')

        # deploy the chain code
        tx_context_dep = create_tx_context(org1_admin,
                                           crypto,
                                           tran_prop_req_dep)

        # create a channel
        request = build_channel_request(self.client,
                                        self.channel_tx,
                                        self.channel_name)

        loop.run_until_complete(self.client._create_or_update_channel(request))

        # join channel
        join_req = loop.run_until_complete(
            build_join_channel_req(org1, channel, self.client))
        responses = channel.join_channel(join_req)
        res = loop.run_until_complete(asyncio.gather(*responses))
        self.assertTrue(all([x.response.status == 200 for x in res]))

        responses, proposal, header = self.client.send_install_proposal(
            tx_context_in, [peer])
        loop.run_until_complete(asyncio.gather(*responses))

        # send the transaction to the channel
        responses, proposal, header = channel.send_instantiate_proposal(
            tx_context_dep, [peer])
        res = loop.run_until_complete(asyncio.gather(*responses))
        tran_req = build_tx_req((res, proposal, header))

        tx_context = create_tx_context(org1_admin,
                                       crypto,
                                       TXProposalRequest())

        responses = loop.run_until_complete(get_stream_result(
            send_transaction(channel.orderers, tran_req, tx_context)))

        self.assertTrue(all([x.status == 200 for x in responses]))



import logging

from numbers import Number

_logger = logging.getLogger(__name__)


class IdentityService(object):

    def __init__(self, client):
        self._client = client

    def create(self, registrar, enrollmentID, enrollmentSecret=None, role=None,
               affiliation=None, maxEnrollments=1, attrs=None):
        """Create a new identity with the Fabric CA server.
        An enrollment secret is returned which can then be used, along with
        the enrollment ID, to enroll a new identity.
        The caller must have `hf.Registrar` authority.

        :param registrar: registrar
        :param enrollmentID: enrollmentID ID which will be used for enrollment
        :param enrollmentSecret: enrollmentSecret Optional enrollment secret to set for the registered user.
        If not provided, the server will generate one.
        When not including, use a null for this parameter.(Default value = None)
        :param role: Optional type of role for this user.
        When not including, use a null for this parameter. (Default value = None)
        :param affiliation: Affiliation with which this user will be associated (Default value = None)
        :param maxEnrollments: The maximum number of times the user is permitted to enroll (Default value = 1)
        :param attrs: Array of key/value attributes to assign to the user. (Default value = None)
        :return: The enrollment secret to use when this user enrolls
        :raises RequestException: errors in requests.exceptions
        :raises ValueError: Failed response, json parse error, args missing
        """

        # TODO, default should be equal to registrar
        # https://hyperledger-fabric-ca.readthedocs.io/en/latest/users-guide.ht
        # ml#registering-a-new-identity
        if affiliation is None:
            affiliation = ''

        if not enrollmentID:
            msg = 'Missing required parameter. \'enrollmentID\' is required.'
            raise ValueError(msg)
        if not isinstance(maxEnrollments, Number):
            msg = 'Wrong parameter. \'maxEnrollments\' must be a' \
                  ' number'
            raise ValueError(msg)

        req = {
            'id': enrollmentID,
            'affiliation': affiliation,
            'max_enrollments': maxEnrollments,
        }

        if role:
            req['type'] = role

        if attrs:
            req['attrs'] = attrs

        if isinstance(enrollmentSecret, str) and len(enrollmentSecret):
            req['secret'] = enrollmentSecret

        authorization = self._client.generateAuthToken(req, registrar)
        headers = {'Authorization': authorization}
        verify = self._client._ca_certs_path
        res, st = self._client._send_ca_post(path='identities',
                                             json=req,
                                             headers=headers,
                                             verify=verify)

        _logger.debug("Response status {0}".format(st))
        _logger.debug("Raw response json {0}".format(res))

        if res['success']:
            return res['result']['secret']
        else:
            raise ValueError("Registering failed with errors {0}"
                             .format(res['errors']))

    def getOne(self, enrollmentID, registrar):

        if not isinstance(enrollmentID, str):
            raise ValueError('argument "enrollmentID" is not a valid string')

        path = 'identities/' + enrollmentID + '?ca=' + self._client._ca_name
        authorization = self._client.generateAuthToken(None, registrar)
        headers = {'Authorization': authorization}
        verify = self._client._ca_certs_path
        res, st = self._client._send_ca_get(path,
                                            headers=headers,
                                            verify=verify)

        _logger.debug("Response status {0}".format(st))
        _logger.debug("Raw response json {0}".format(res))

        return res

    def getAll(self, registrar):

        path = 'identities?ca=' + self._client._ca_name
        authorization = self._client.generateAuthToken(None, registrar)
        headers = {'Authorization': authorization}
        verify = self._client._ca_certs_path
        res, st = self._client._send_ca_get(path,
                                            headers=headers,
                                            verify=verify)

        _logger.debug("Response status {0}".format(st))
        _logger.debug("Raw response json {0}".format(res))

        return res

    def delete(self, enrollmentID, registrar, force=False):

        if not isinstance(enrollmentID, str):
            raise ValueError('argument "enrollmentID" is not a valid string')

        path = 'identities/' + enrollmentID
        if force is True:
            path += '?force=true'

        authorization = self._client.generateAuthToken(None, registrar)
        headers = {'Authorization': authorization}
        verify = self._client._ca_certs_path
        res, st = self._client._send_ca_delete(path,
                                               headers=headers,
                                               verify=verify)

        _logger.debug("Response status {0}".format(st))
        _logger.debug("Raw response json {0}".format(res))

        return res

    def update(self, enrollmentID, registrar, type=None, affiliation=None,
               maxEnrollments=None, attrs=None, enrollmentSecret=None,
               caname=None):

        if not isinstance(enrollmentID, str):
            raise ValueError('argument "enrollmentID" is not a valid string')

        path = 'identities/' + enrollmentID

        req = {}
        if type:
            req['type'] = type
        if affiliation:
            req['affiliation'] = affiliation
        if isinstance(maxEnrollments, Number):
            req['max_enrollments'] = maxEnrollments
        if attrs:
            req['attrs'] = attrs
        if enrollmentSecret:
            req['secret'] = enrollmentSecret
        if caname:
            req['caname'] = caname

        authorization = self._client.generateAuthToken(req, registrar)
        headers = {'Authorization': authorization}
        verify = self._client._ca_certs_path
        res, st = self._client._send_ca_update(path,
                                               json=req,
                                               headers=headers,
                                               verify=verify)

        _logger.debug("Response status {0}".format(st))
        _logger.debug("Raw response json {0}".format(res))

        return res



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hfc/protos/peer/admin.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from hfc.protos.common import common_pb2 as hfc_dot_protos_dot_common_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hfc/protos/peer/admin.proto',
  package='protos',
  syntax='proto3',
  serialized_options=_b('\n\"org.hyperledger.fabric.protos.peerB\014AdminPackageZ)github.com/hyperledger/fabric/protos/peer'),
  serialized_pb=_b('\n\x1bhfc/protos/peer/admin.proto\x12\x06protos\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1ehfc/protos/common/common.proto\"\x9a\x01\n\x0cServerStatus\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x1f.protos.ServerStatus.StatusCode\"Y\n\nStatusCode\x12\r\n\tUNDEFINED\x10\x00\x12\x0b\n\x07STARTED\x10\x01\x12\x0b\n\x07STOPPED\x10\x02\x12\n\n\x06PAUSED\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\x0b\n\x07UNKNOWN\x10\x05\"8\n\x0fLogLevelRequest\x12\x12\n\nlog_module\x18\x01 \x01(\t\x12\x11\n\tlog_level\x18\x02 \x01(\t\"9\n\x10LogLevelResponse\x12\x12\n\nlog_module\x18\x01 \x01(\t\x12\x11\n\tlog_level\x18\x02 \x01(\t\"\"\n\x0eLogSpecRequest\x12\x10\n\x08log_spec\x18\x01 \x01(\t\"2\n\x0fLogSpecResponse\x12\x10\n\x08log_spec\x18\x01 \x01(\t\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"t\n\x0e\x41\x64minOperation\x12)\n\x06logReq\x18\x01 \x01(\x0b\x32\x17.protos.LogLevelRequestH\x00\x12,\n\nlogSpecReq\x18\x02 \x01(\x0b\x32\x16.protos.LogSpecRequestH\x00\x42\t\n\x07\x63ontent2\xb2\x03\n\x05\x41\x64min\x12\x35\n\tGetStatus\x12\x10.common.Envelope\x1a\x14.protos.ServerStatus\"\x00\x12\x37\n\x0bStartServer\x12\x10.common.Envelope\x1a\x14.protos.ServerStatus\"\x00\x12\x41\n\x11GetModuleLogLevel\x12\x10.common.Envelope\x1a\x18.protos.LogLevelResponse\"\x00\x12\x41\n\x11SetModuleLogLevel\x12\x10.common.Envelope\x1a\x18.protos.LogLevelResponse\"\x00\x12=\n\x0fRevertLogLevels\x12\x10.common.Envelope\x1a\x16.google.protobuf.Empty\"\x00\x12\x39\n\nGetLogSpec\x12\x10.common.Envelope\x1a\x17.protos.LogSpecResponse\"\x00\x12\x39\n\nSetLogSpec\x12\x10.common.Envelope\x1a\x17.protos.LogSpecResponse\"\x00\x42]\n\"org.hyperledger.fabric.protos.peerB\x0c\x41\x64minPackageZ)github.com/hyperledger/fabric/protos/peerb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,hfc_dot_protos_dot_common_dot_common__pb2.DESCRIPTOR,])



_SERVERSTATUS_STATUSCODE = _descriptor.EnumDescriptor(
  name='StatusCode',
  full_name='protos.ServerStatus.StatusCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STARTED', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STOPPED', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAUSED', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=166,
  serialized_end=255,
)
_sym_db.RegisterEnumDescriptor(_SERVERSTATUS_STATUSCODE)


_SERVERSTATUS = _descriptor.Descriptor(
  name='ServerStatus',
  full_name='protos.ServerStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='protos.ServerStatus.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SERVERSTATUS_STATUSCODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=255,
)


_LOGLEVELREQUEST = _descriptor.Descriptor(
  name='LogLevelRequest',
  full_name='protos.LogLevelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_module', full_name='protos.LogLevelRequest.log_module', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log_level', full_name='protos.LogLevelRequest.log_level', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=313,
)


_LOGLEVELRESPONSE = _descriptor.Descriptor(
  name='LogLevelResponse',
  full_name='protos.LogLevelResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_module', full_name='protos.LogLevelResponse.log_module', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log_level', full_name='protos.LogLevelResponse.log_level', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=315,
  serialized_end=372,
)


_LOGSPECREQUEST = _descriptor.Descriptor(
  name='LogSpecRequest',
  full_name='protos.LogSpecRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_spec', full_name='protos.LogSpecRequest.log_spec', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=374,
  serialized_end=408,
)


_LOGSPECRESPONSE = _descriptor.Descriptor(
  name='LogSpecResponse',
  full_name='protos.LogSpecResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_spec', full_name='protos.LogSpecResponse.log_spec', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='protos.LogSpecResponse.error', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=410,
  serialized_end=460,
)


_ADMINOPERATION = _descriptor.Descriptor(
  name='AdminOperation',
  full_name='protos.AdminOperation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='logReq', full_name='protos.AdminOperation.logReq', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='logSpecReq', full_name='protos.AdminOperation.logSpecReq', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='content', full_name='protos.AdminOperation.content',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=462,
  serialized_end=578,
)

_SERVERSTATUS.fields_by_name['status'].enum_type = _SERVERSTATUS_STATUSCODE
_SERVERSTATUS_STATUSCODE.containing_type = _SERVERSTATUS
_ADMINOPERATION.fields_by_name['logReq'].message_type = _LOGLEVELREQUEST
_ADMINOPERATION.fields_by_name['logSpecReq'].message_type = _LOGSPECREQUEST
_ADMINOPERATION.oneofs_by_name['content'].fields.append(
  _ADMINOPERATION.fields_by_name['logReq'])
_ADMINOPERATION.fields_by_name['logReq'].containing_oneof = _ADMINOPERATION.oneofs_by_name['content']
_ADMINOPERATION.oneofs_by_name['content'].fields.append(
  _ADMINOPERATION.fields_by_name['logSpecReq'])
_ADMINOPERATION.fields_by_name['logSpecReq'].containing_oneof = _ADMINOPERATION.oneofs_by_name['content']
DESCRIPTOR.message_types_by_name['ServerStatus'] = _SERVERSTATUS
DESCRIPTOR.message_types_by_name['LogLevelRequest'] = _LOGLEVELREQUEST
DESCRIPTOR.message_types_by_name['LogLevelResponse'] = _LOGLEVELRESPONSE
DESCRIPTOR.message_types_by_name['LogSpecRequest'] = _LOGSPECREQUEST
DESCRIPTOR.message_types_by_name['LogSpecResponse'] = _LOGSPECRESPONSE
DESCRIPTOR.message_types_by_name['AdminOperation'] = _ADMINOPERATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServerStatus = _reflection.GeneratedProtocolMessageType('ServerStatus', (_message.Message,), dict(
  DESCRIPTOR = _SERVERSTATUS,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.ServerStatus)
  ))
_sym_db.RegisterMessage(ServerStatus)

LogLevelRequest = _reflection.GeneratedProtocolMessageType('LogLevelRequest', (_message.Message,), dict(
  DESCRIPTOR = _LOGLEVELREQUEST,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.LogLevelRequest)
  ))
_sym_db.RegisterMessage(LogLevelRequest)

LogLevelResponse = _reflection.GeneratedProtocolMessageType('LogLevelResponse', (_message.Message,), dict(
  DESCRIPTOR = _LOGLEVELRESPONSE,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.LogLevelResponse)
  ))
_sym_db.RegisterMessage(LogLevelResponse)

LogSpecRequest = _reflection.GeneratedProtocolMessageType('LogSpecRequest', (_message.Message,), dict(
  DESCRIPTOR = _LOGSPECREQUEST,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.LogSpecRequest)
  ))
_sym_db.RegisterMessage(LogSpecRequest)

LogSpecResponse = _reflection.GeneratedProtocolMessageType('LogSpecResponse', (_message.Message,), dict(
  DESCRIPTOR = _LOGSPECRESPONSE,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.LogSpecResponse)
  ))
_sym_db.RegisterMessage(LogSpecResponse)

AdminOperation = _reflection.GeneratedProtocolMessageType('AdminOperation', (_message.Message,), dict(
  DESCRIPTOR = _ADMINOPERATION,
  __module__ = 'hfc.protos.peer.admin_pb2'
  # @@protoc_insertion_point(class_scope:protos.AdminOperation)
  ))
_sym_db.RegisterMessage(AdminOperation)


DESCRIPTOR._options = None

_ADMIN = _descriptor.ServiceDescriptor(
  name='Admin',
  full_name='protos.Admin',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=581,
  serialized_end=1015,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetStatus',
    full_name='protos.Admin.GetStatus',
    index=0,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_SERVERSTATUS,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StartServer',
    full_name='protos.Admin.StartServer',
    index=1,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_SERVERSTATUS,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetModuleLogLevel',
    full_name='protos.Admin.GetModuleLogLevel',
    index=2,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_LOGLEVELRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SetModuleLogLevel',
    full_name='protos.Admin.SetModuleLogLevel',
    index=3,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_LOGLEVELRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RevertLogLevels',
    full_name='protos.Admin.RevertLogLevels',
    index=4,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetLogSpec',
    full_name='protos.Admin.GetLogSpec',
    index=5,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_LOGSPECRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SetLogSpec',
    full_name='protos.Admin.SetLogSpec',
    index=6,
    containing_service=None,
    input_type=hfc_dot_protos_dot_common_dot_common__pb2._ENVELOPE,
    output_type=_LOGSPECRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_ADMIN)

DESCRIPTOR.services_by_name['Admin'] = _ADMIN

# @@protoc_insertion_point(module_scope)



# Copyright sudheesh.info 2018 All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import binascii
import datetime
from base64 import b64encode
from datetime import timezone

# Import required Peer Protos
from hfc.protos.peer import chaincode_event_pb2
from hfc.protos.peer import transaction_pb2
from hfc.protos.peer import proposal_pb2
from hfc.protos.peer import proposal_response_pb2
from hfc.protos.peer import configuration_pb2 as peer_configuration_pb2
from hfc.protos.peer import events_pb2

# Import required MSP Protos
from hfc.protos.msp import msp_principal_pb2
from hfc.protos.msp import msp_config_pb2
from hfc.protos.msp import identities_pb2

# Import required Common Protos
from hfc.protos.common import common_pb2
from hfc.protos.common import configtx_pb2
from hfc.protos.common import policies_pb2
from hfc.protos.common import configuration_pb2 as common_configuration_pb2

# Import required Orderer Protos
from hfc.protos.orderer import configuration_pb2 as orderer_configuration_pb2

# Import required Ledger Protos
from hfc.protos.ledger.rwset import rwset_pb2
from hfc.protos.ledger.rwset.kvrwset import kv_rwset_pb2

# Import required Gossip Protos
from hfc.protos.gossip import message_pb2

_logger = logging.getLogger(__name__ + ".block_decoder")


class BlockDecoder(object):
    """An object of a fully decoded protobuf message "Block" """

    @staticmethod
    def decode(block_bytes):
        """Constructs a JSON Object containing all decoded values from
        protobuf encoded `Block` bytes.

        :param block_bytes: Block instance
        :return: Dictionary containing decoded Block instance.
        :type block_bytes: bytes
        """
        block = {}
        try:
            proto_block = common_pb2.Block()
            proto_block.ParseFromString(block_bytes)
            block['header'] = decode_block_header(proto_block.header)
            block['data'] = decode_block_data(proto_block.data, True)
            block['metadata'] = decode_block_metadata(proto_block.metadata)
        except Exception as e:
            raise ValueError("BlockDecoder :: decode failed", e)
        return block

    @staticmethod
    def decode_transaction(processed_tx_bytes):
        """Decodes a transaction proto and constructs a deserialized object

        :param processed_tx_bytes: Binary content of tx
        :return: Dictionary containing tx block information
        :raises ValueError: If data is not passed to the method
        """
        if not processed_tx_bytes:
            raise ValueError("BlockDecoder :: decode_transaction \
                doesnot have processed transaction bytes")
        processed_tx = {}
        pr_processed_tx = transaction_pb2.ProcessedTransaction()
        pr_processed_tx.ParseFromString(processed_tx_bytes)
        if pr_processed_tx:
            processed_tx['validation_code'] = \
                pr_processed_tx.validationCode
            processed_tx['transaction_envelope'] = \
                decode_block_data_envelope(pr_processed_tx.transactionEnvelope)
        return processed_tx


class FilteredBlockDecoder(object):
    """An object of a fully decoded protobuf message "FilteredBlock" """

    @staticmethod
    def decode(block_bytes):
        """Constructs a JSON Object containing all decoded values from
        protobuf encoded `FilteredBlock` bytes.

        :param block_bytes: FilteredBlock instance
        :return: Dictionary containing decoded Filtered Block instance.
        :type block_bytes: bytes
        """

        filtered_block = {}

        try:
            proto_block = events_pb2.FilteredBlock()
            proto_block.ParseFromString(block_bytes)
            filtered_block['channel_id'] = proto_block.channel_id
            filtered_block['number'] = proto_block.number
            filtered_block['filtered_transactions'] = []
            fts = proto_block.filtered_transactions

            for ft in fts:
                code = tx_validation_code.get(ft.tx_validation_code,
                                              'UNKNOWN_VALIDATION_CODE')
                ft_decoded = {
                    'txid': ft.txid,
                    'type': HeaderType.convert_to_string(ft.type),
                    'tx_validation_code': code
                }

                if hasattr(ft, 'transaction_actions'):
                    tx_a = {'chaincode_actions': []}
                    for ca in ft.transaction_actions.chaincode_actions:
                        cce = decode_chaincode_events(
                            ca.chaincode_event.SerializeToString())
                        tx_a['chaincode_actions'].append({
                            'chaincode_event': cce
                        })
                    ft_decoded['transaction_actions'] = tx_a
                filtered_block['filtered_transactions'].append(ft_decoded)

        except Exception as e:
            raise ValueError("FilteredBlockDecoder :: decode failed", e)
        return filtered_block


tx_validation_code = {
    0: 'VALID',
    1: 'NIL_ENVELOPE',
    2: 'BAD_PAYLOAD',
    3: 'BAD_COMMON_HEADER',
    4: 'BAD_CREATOR_SIGNATURE',
    5: 'INVALID_ENDORSER_TRANSACTION',
    6: 'INVALID_CONFIG_TRANSACTION',
    7: 'UNSUPPORTED_TX_PAYLOAD',
    8: 'BAD_PROPOSAL_TXID',
    9: 'DUPLICATE_TXID',
    10: 'ENDORSEMENT_POLICY_FAILURE',
    11: 'MVCC_READ_CONFLICT',
    12: 'PHANTOM_READ_CONFLICT',
    13: 'UNKNOWN_TX_TYPE',
    14: 'TARGET_CHAIN_NOT_FOUND',
    15: 'MARSHAL_TX_ERROR',
    16: 'NIL_TXACTION',
    17: 'EXPIRED_CHAINCODE',
    18: 'CHAINCODE_VERSION_CONFLICT',
    19: 'BAD_HEADER_EXTENSION',
    20: 'BAD_CHANNEL_HEADER',
    21: 'BAD_RESPONSE_PAYLOAD',
    22: 'BAD_RWSET',
    23: 'ILLEGAL_WRITESET',
    24: 'INVALID_WRITESET',
    254: 'NOT_VALIDATED',
    255: 'INVALID_OTHER_REASON',
}

type_as_string = {
    0: 'MESSAGE',  # Used for messages which are signed but opaque
    1: 'CONFIG',  # Used for messages which express the channel config
    2: 'CONFIG_UPDATE',  # Used for transactions that update the channel config
    3: 'ENDORSER_TRANSACTION',  # Used to submit endorser based transactions
    4: 'ORDERER_TRANSACTION',  # Used internally by the orderer for management
    5: 'DELIVER_SEEK_INFO',  # Used to instruct the Deliver API to seek
    6: 'CHAINCODE_PACKAGE'  # Used to packaging chaincode artifacts for install
}

implicit_metapolicy_rule = ['ANY', 'ALL', 'MAJORITY']

policy_policy_type = ['UNKNOWN', 'SIGNATURE', 'MSP', 'IMPLICIT_META']


class HeaderType(object):
    """HeaderType class having decodePayload and convertToString methods"""

    @staticmethod
    def convert_to_string(type_value):
        return type_as_string.get(type_value, 'UNKNOWN_TYPE')

    @staticmethod
    def decode_payload_based_on_type(proto_data, type_value):
        result = None
        if type_value == 1:
            result = decode_config_envelope(proto_data)
        elif type_value == 2:
            result = decode_config_update_envelope(proto_data)
        elif type_value == 3:
            result = decode_endorser_transaction(proto_data)
        else:
            msg = f'HeaderType :: decode_payload found a header type of' \
                f' {type_value} :: {HeaderType.convert_to_string(type_value)}'
            _logger.debug(msg)
            result = {}
        return result


def decode_block_header(proto_block_header):
    """Decodes the header of Block

    :param proto_block_header: Block Header proto
    :return: Decoded BlockHeader inside Block instance.
    :type proto_block_header: str
    """
    block_header = {}
    block_header['number'] = proto_block_header.number
    block_header['previous_hash'] = \
        binascii.b2a_hex(proto_block_header.previous_hash)
    block_header['data_hash'] = binascii.b2a_hex(proto_block_header.data_hash)
    return block_header


def decode_block_data(proto_block_data, not_proto=False):
    """Decodes the data of Block.

    :param proto_block_data: Block Data proto.
    :type proto_block_data: str
    :param not_proto: Boolean for if proto.
    :return: deserialized block_data (Default value = False)
    :type not_proto: bool
    """
    data = {}
    data['data'] = []
    for i in proto_block_data.data:
        proto_envelope = None
        if not_proto:
            proto_envelope = common_pb2.Envelope()
            proto_envelope.ParseFromString(i)
        if proto_envelope:
            envelope = decode_block_data_envelope(proto_envelope)
            data['data'].append(envelope)
    return data


def decode_block_metadata(proto_block_metadata):
    """Decodes block metadata from block

    :param proto_block_metadata: Block metadata proto content
    :return: deserialized metadata contents
    :type proto_block_metadata: bytes
    """
    metadata = {}
    metadata['metadata'] = []
    if proto_block_metadata and proto_block_metadata.metadata:
        signatures = decode_metadata_signatures(
            proto_block_metadata.metadata[common_pb2.SIGNATURES])
        metadata['metadata'].append(signatures)

        last_config = decode_last_config_sequence_number(
            proto_block_metadata.metadata[common_pb2.LAST_CONFIG])
        metadata['metadata'].append(last_config)

        transaction_filter = decode_transaction_filter(
            proto_block_metadata.metadata[common_pb2.TRANSACTIONS_FILTER])
        metadata['metadata'].append(transaction_filter)

    return metadata


def decode_block_data_envelope(proto_envelope):
    """Decodes the envelope contents of Block

    :param proto_envelope: Envelope proto
    :return: deserialized block envelope
    :type proto_envelope: str
    """
    envelope = {}
    envelope['signature'] = proto_envelope.signature
    envelope['payload'] = {}
    proto_payload = common_pb2.Payload()
    proto_payload.ParseFromString(proto_envelope.payload)
    envelope['payload']['header'] = decode_header(proto_payload.header)
    envelope['payload']['data'] = \
        HeaderType.decode_payload_based_on_type(
            proto_payload.data,
            envelope['payload']['header']['channel_header']['type'])
    envelope['payload']['header']['channel_header']['type_string'] = \
        HeaderType.convert_to_string(
            envelope['payload']['header']['channel_header']['type'])
    return envelope


def decode_header(proto_header):
    """Decodes the Payload header in envelope

    :param proto_header: Envelope Payload
    :return: deserialized envelope header
    :type proto_header: str
    """
    header = {}
    header['channel_header'] = \
        decode_channel_header(proto_header.channel_header)
    header['signature_header'] = \
        decode_signature_header(proto_header.signature_header)
    return header


def decode_channel_header(header_bytes):
    """Decodes channel header for Payload channel header

    :param header_bytes: Bytes channel header
    :return: deserialized payload channel_header
    :type header_bytes: str
    """
    channel_header = {}
    proto_channel_header = common_pb2.ChannelHeader()
    proto_channel_header.ParseFromString(header_bytes)
    channel_header['type'] = proto_channel_header.type
    channel_header['version'] = decode_version(proto_channel_header.version)
    channel_header['timestamp'] = \
        timestamp_to_date(proto_channel_header.timestamp)
    channel_header['channel_id'] = proto_channel_header.channel_id
    channel_header['tx_id'] = proto_channel_header.tx_id
    channel_header['epoch'] = proto_channel_header.epoch
    channel_header['extension'] = proto_channel_header.extension
    return channel_header


def timestamp_to_date(timestamp):
    """Converts timestamp to current date

    :param timestamp: Timestamp value
    :return: String formatted date in %Y-%m-%d %H:%M:%S
    """
    if not timestamp:
        return None
    # WARNING: this will break on Windows because of the fromtimestamp()
    # restriction of values by C `localtime()` or `gmtime()` calls.
    millis = timestamp.seconds * 1000 + timestamp.nanos / 1000000
    date = datetime.datetime.fromtimestamp(millis / 1e3, tz=timezone.utc)
    return date.strftime("%Y-%m-%d %H:%M:%S")


def decode_version(version_long):
    """Takes version proto object and returns version

    :param version_long: version proto object
    :return: integer value of version_long
    """
    return int(version_long)


def decode_signature_header(signature_header_bytes):
    """Decode signature header

    :param signature_header_bytes: signature header bytes
    :return: deserialized signature_header
    """
    signature_header = {}
    proto_signature_header = common_pb2.SignatureHeader()
    proto_signature_header.ParseFromString(signature_header_bytes)
    signature_header['creator'] = \
        decode_identity(proto_signature_header.creator)
    signature_header['nonce'] = \
        binascii.b2a_hex(proto_signature_header.nonce)
    return signature_header


def decode_identity(id_bytes):
    """Decodes identity

    :param id_bytes: byte of identity
    :return: deserialized identity
    """
    identity = {}
    try:
        proto_identity = identities_pb2.SerializedIdentity()
        proto_identity.ParseFromString(id_bytes)
        identity['mspid'] = proto_identity.mspid
        identity['id_bytes'] = proto_identity.id_bytes.decode()
    except Exception as e:
        raise ValueError("BlockDecoder :: decode_identiy failed", e)
    return identity


def decode_metadata_signatures(metadata_bytes):
    """Decodes metadata signature from bytes

    :param metadata_bytes: Metadata object proto
    :return: deserialized Metadata blocks
    :type metadata_bytes: str
    """
    metadata = {}
    proto_metadata = common_pb2.Metadata()
    proto_metadata.ParseFromString(metadata_bytes)
    metadata['value'] = proto_metadata.value
    metadata['signatures'] = \
        decode_metadata_value_signatures(proto_metadata.signatures)
    return metadata


def decode_metadata_value_signatures(proto_meta_signatures):
    """Decodes all signatures in metadata values

    :param proto_meta_signatures: List of value objects
    :return: deserialized list of signatures from metadata values
    :type proto_meta_signatures: list(str
    """
    signatures = []
    if proto_meta_signatures:
        for signature in proto_meta_signatures:
            metadata_signature = {}
            metadata_signature['signature_header'] = \
                decode_signature_header(signature.signature_header)
            metadata_signature['signature'] = signature.signature
            signatures.append(metadata_signature)
    return signatures


def decode_last_config_sequence_number(metadata_bytes):
    """Decodes last configuration and index for sequence number

    :param metadata_bytes: encoded content for sequence number
    :return: deserialized dictionary of config sequence number
    :type metadata_bytes: str
    """
    last_config = {
        'value': {
            'index': 0,
            'signatures': []
        }
    }
    if metadata_bytes:
        proto_metadata = common_pb2.Metadata()
        proto_metadata.ParseFromString(metadata_bytes)
        proto_last_config = common_pb2.LastConfig()
        proto_last_config.ParseFromString(proto_metadata.value)
        last_config['value']['index'] = proto_last_config.index
        last_config['signatures'] = \
            decode_metadata_value_signatures(proto_metadata.signatures)
    return last_config


def decode_transaction_filter(metadata_bytes):
    """Decodes transaction filter from metadata bytes

    :param metadata_bytes: Encoded list of transaction filters
    :return: decoded transaction_filter list
    :type metadata_bytes: str
    """
    transaction_filter = []
    if not metadata_bytes:
        return None

    for i in metadata_bytes:
        transaction_filter.append(int(i))
    return transaction_filter


def decode_endorser_transaction(trans_bytes):
    """Decodes

    :param trans_bytes: Serialized endorser transaction bytes
    :return: deserialized dictionary of endorser transaction data
    """
    data = {}
    if trans_bytes:
        transaction = transaction_pb2.Transaction()
        transaction.ParseFromString(trans_bytes)
        data['actions'] = []
        if transaction and transaction.actions:
            for tx_action in transaction.actions:
                action = {}
                action['header'] = \
                    decode_signature_header(tx_action.header)
                action['payload'] = \
                    decode_chaincode_action_payload(tx_action.payload)
                data['actions'].append(action)
    return data


def decode_config_envelope(config_envelope_bytes):
    """Decodes configuration envelope

    :param config_envelope_bytes: byte of config envelope
    :return: deserialized config envelope
    """
    config_envelope = {}
    proto_config_envelope = configtx_pb2.ConfigEnvelope()
    proto_config_envelope.ParseFromString(config_envelope_bytes)
    config_envelope['config'] = decode_config(proto_config_envelope.config)
    config_envelope['last_update'] = {}
    proto_last_update = proto_config_envelope.last_update
    if proto_last_update:
        config_envelope['last_update']['payload'] = {}
        proto_payload = common_pb2.Payload()
        proto_payload.ParseFromString(proto_last_update.payload)
        config_envelope['last_update']['payload']['header'] = \
            decode_header(proto_payload.header)
        config_envelope['last_update']['payload']['data'] = \
            decode_config_update_envelope(proto_payload.data)
        config_envelope['last_update']['signature'] = \
            proto_last_update.signature
    return config_envelope


def decode_config(proto_config):
    """Decodes configuration from config envelope

    :param proto_config: Config value
    :type proto_config: bytes
    :return: deserialized config
    """
    config = {}
    config['sequence'] = str(proto_config.sequence)
    config['channel_group'] = decode_config_group(proto_config.channel_group)
    # config['type'] = proto_config.type
    # TODO: getType() equivalent
    return config


def decode_config_update_envelope(config_update_envelope_bytes):
    """Decode config update envelope

    :param config_update_envelope_bytes: Bytes of update envelope
    :type config_update_envelope_bytes: str
    :return: deserialized config update envelope signatures
    """
    config_update_envelope = {}
    proto_config_update_envelope = configtx_pb2.ConfigUpdateEnvelope()
    proto_config_update_envelope.ParseFromString(config_update_envelope_bytes)
    config_update_envelope['config_update'] = \
        decode_config_update(proto_config_update_envelope.config_update)
    signatures = []
    for signature in proto_config_update_envelope.signatures:
        proto_config_signature = signature
        config_signature = decode_config_signature(proto_config_signature)
        signatures.append(config_signature)
    config_update_envelope['signatures'] = signatures
    return config_update_envelope


def decode_config_update(config_update_bytes):
    """Decodes update bytes in configuration

    :param config_update_bytes: Bytes
    :type config_update_bytes: str
    :return: deserialized configuration update
    """
    config_update = {}
    proto_config_update = configtx_pb2.ConfigUpdate()
    proto_config_update.ParseFromString(config_update_bytes)
    config_update['channel_id'] = proto_config_update.channel_id
    config_update['read_set'] = \
        decode_config_group(proto_config_update.read_set)
    config_update['write_set'] = \
        decode_config_group(proto_config_update.write_set)
    # config_update['type'] = proto_config_update TODO: getType() equivalent
    return config_update


def decode_config_groups(config_group_map):
    """Decodes configuration groups inside ConfigGroup

    :param config_group_map: Serialized ConfigGroup.groups object
    :type config_group_map: str
    :return: map of configuration groups.
    """
    config_groups = {}
    keys = config_group_map.keys()
    for key in keys:
        config_groups[key] = decode_config_group(config_group_map[key])
    return config_groups


def decode_config_group(proto_config_group):
    """Decodes configuration group from config protos

    :param proto_config_group: serialized ConfigGroup() object
    :type proto_config_group: str
    :return: deserialized config_groups dictionary
    """

    if not proto_config_group:
        return None
    config_group = {}
    config_group['version'] = decode_version(proto_config_group.version)
    config_group['groups'] = decode_config_groups(proto_config_group.groups)
    config_group['values'] = decode_config_values(proto_config_group.values)
    config_group['policies'] = \
        decode_config_policies(proto_config_group.policies)
    config_group['mod_policy'] = proto_config_group.mod_policy
    return config_group


def decode_config_values(config_value_map):
    """Decodes configuration values inside each configuration key

    :param config_value_map: Serialized values map for each config key
    :type config_value_map: str
    :return: map of configuration values for each key
    """
    config_values = {}
    keys = config_value_map.keys()
    for key in keys:
        config_values[key] = decode_config_value(config_value_map[key], key)
    return config_values


def decode_config_value(proto_config_value, key):
    """Decodes ConfigValue from map with a given key

    :param proto_config_value: A bytes string of config_value
    :type proto_config_value: str
    :param key: Map key for the configuration value
    :type key: str
    :return: Dictionary of configuration value deserialized
    """
    config_value_key = key
    config_value = {}
    config_value['version'] = decode_version(proto_config_value.version)
    config_value['mod_policy'] = proto_config_value.mod_policy
    config_value['value'] = {}
    if config_value_key == 'AnchorPeers':
        anchor_peers = []
        proto_anchor_peers = peer_configuration_pb2.AnchorPeers()
        proto_anchor_peers.ParseFromString(proto_config_value.value)
        if proto_anchor_peers and proto_anchor_peers.anchor_peers:
            for peer in proto_anchor_peers.anchor_peers:
                anchor_peer = {}
                anchor_peer['host'] = peer.host
                anchor_peer['port'] = peer.port
                anchor_peers.append(anchor_peer)
            config_value['value']['anchor_peers'] = anchor_peers
    elif config_value_key == 'MSP':
        msp_config = {}
        proto_msp_config = msp_config_pb2.MSPConfig()
        proto_msp_config.ParseFromString(proto_config_value.value)
        if proto_msp_config.type == 0:
            msp_config = decode_fabric_MSP_config(proto_msp_config.config)
        config_value['value']['type'] = proto_msp_config.type
        config_value['value']['config'] = msp_config
    elif config_value_key == 'ConsensusType':
        proto_consensus_type = orderer_configuration_pb2.ConsensusType()
        proto_consensus_type.ParseFromString(proto_config_value.value)
        config_value['value']['type'] = proto_consensus_type.type
    elif config_value_key == 'BatchSize':
        proto_batch_size = orderer_configuration_pb2.BatchSize()
        proto_batch_size.ParseFromString(proto_config_value.value)
        config_value['value']['max_message_count'] = \
            proto_batch_size.max_message_count
        config_value['value']['absolute_max_bytes'] = \
            proto_batch_size.absolute_max_bytes
        config_value['value']['preferred_max_bytes'] = \
            proto_batch_size.preferred_max_bytes
    elif config_value_key == 'BatchTimeout':
        proto_batch_timeout = orderer_configuration_pb2.BatchTimeout()
        proto_batch_timeout.ParseFromString(proto_config_value.value)
        config_value['value']['timeout'] = proto_batch_timeout.timeout
    elif config_value_key == 'ChannelRestrictions':
        proto_channel_restrictions = \
            orderer_configuration_pb2.ChannelRestrictions()
        proto_channel_restrictions.ParseFromString(proto_config_value.value)
        config_value['value']['max_count'] = \
            str(proto_channel_restrictions.max_count)
    elif config_value_key == 'Consortium':
        consortium_name = common_configuration_pb2.Consortium()
        consortium_name.ParseFromString(proto_config_value.value)
        config_value['value']['name'] = consortium_name.name
    elif config_value_key == 'HashingAlgorithm':
        proto_hashing_algorithm = common_configuration_pb2.HashingAlgorithm()
        proto_hashing_algorithm.ParseFromString(proto_config_value.value)
        config_value['value']['name'] = proto_hashing_algorithm.name
    elif config_value_key == 'BlockDataHashingStructure':
        proto_blockdata_hashing_structure = \
            common_configuration_pb2.BlockDataHashingStructure()
        proto_blockdata_hashing_structure.ParseFromString(
            proto_config_value.value)
        config_value['value']['width'] = \
            proto_blockdata_hashing_structure.width
    elif config_value_key == 'OrdererAddresses':
        orderer_addresses = common_configuration_pb2.OrdererAddresses()
        orderer_addresses.ParseFromString(proto_config_value.value)
        addresses = []
        proto_addresses = orderer_addresses.addresses
        if proto_addresses:
            for address in proto_addresses:
                addresses.append(address)
            config_value['value']['addresses'] = addresses
    else:
        pass
    return config_value


def decode_config_policies(config_policy_map):
    """Decodes list of configuration policies

    :param config_policy_map: Serialized list of configuration policies
    :type config_policy_map: str
    :return: deserialized map of config policies.
    """
    config_policies = {}
    keys = config_policy_map.keys()
    for key in keys:
        config_policies[key] = decode_config_policy(config_policy_map[key])
    return config_policies


def decode_config_policy(proto_config_policy):
    """Decodes config policy based on type of policy

    :param proto_config_policy: Configuration policy bytes
    :return: deserialized config_policy based on policy type.
    """
    config_policy = {}
    config_policy['version'] = decode_version(proto_config_policy.version)
    config_policy['mod_policy'] = proto_config_policy.mod_policy
    config_policy['policy'] = {}
    if proto_config_policy.policy:
        config_policy['policy']['type'] = proto_config_policy.policy.type
        if (proto_config_policy.policy.type == policies_pb2.Policy.SIGNATURE):
            config_policy['policy']['value'] = \
                decode_signature_policy_envelope(
                    proto_config_policy.policy.value)
        elif (proto_config_policy.policy.type == policies_pb2.Policy.MSP):
            proto_msp = policies_pb2.Policy()
            proto_msp.ParseFromString(proto_config_policy.policy.value)
        elif (proto_config_policy.policy.type ==
              policies_pb2.Policy.IMPLICIT_META):
            config_policy['policy']['value'] = \
                decode_implicit_meta_policy(proto_config_policy.policy.value)
        elif (proto_config_policy.policy.type == policies_pb2.Policy.UNKNOWN):
            config_policy['policy']['value'] = 'Unknown'
        else:
            raise ValueError("Unknown policy type")
    return config_policy


def decode_implicit_meta_policy(implicit_meta_policy_bytes):
    """Decodes implicit meta policy in a policy

    :param implicit_meta_policy_bytes: Bytes of implicit meta policy
    :type implicit_meta_policy_bytes: str
    :return: deserialized implicit_meta_policy value.
    """
    implicit_meta_policy = {}
    proto_implicit_meta_policy = policies_pb2.ImplicitMetaPolicy()
    proto_implicit_meta_policy.ParseFromString(implicit_meta_policy_bytes)
    implicit_meta_policy['sub_policy'] = \
        proto_implicit_meta_policy.sub_policy
    implicit_meta_policy['rule'] = \
        implicit_metapolicy_rule[proto_implicit_meta_policy.rule]
    return implicit_meta_policy


def decode_signature_policy_envelope(signature_policy_envelope_bytes):
    """Decodes signature policy envelope bytes

    :param signature_policy_envelope_bytes: Serialized signature envelope
    :type signature_policy_envelope_bytes: str
    :return: deserialized signature policy envelope contents.
    """
    signature_policy_envelope = {}
    proto_signature_policy_envelope = policies_pb2.SignaturePolicyEnvelope()
    proto_signature_policy_envelope.ParseFromString(
        signature_policy_envelope_bytes)
    signature_policy_envelope['version'] = \
        decode_version(proto_signature_policy_envelope.version)
    signature_policy_envelope['rule'] = \
        decode_signature_policy(proto_signature_policy_envelope.rule)
    identities = []
    proto_identities = proto_signature_policy_envelope.identities
    if proto_identities:
        for identity in proto_identities:
            msp_principal = decode_MSP_principal(identity)
            identities.append(msp_principal)
    signature_policy_envelope['identities'] = identities
    return signature_policy_envelope


def decode_signature_policy(proto_signature_policy):
    """Decodes signature policy based on field

    :param proto_signature_policy: Object of SignaturePolicy()
    :return: deserialized signature policy after decoding based on field.
    """
    signature_policy = {}
    if proto_signature_policy.HasField('n_out_of'):
        signature_policy['n_out_of'] = {}
        signature_policy['n_out_of']['n'] = proto_signature_policy.n_out_of.n
        signature_policy['n_out_of']['rules'] = []
        for rule in proto_signature_policy.n_out_of.rules:
            proto_policy = rule
            policy = decode_signature_policy(proto_policy)
            signature_policy['n_out_of']['rules'].append(policy)
    elif proto_signature_policy.HasField('signed_by'):
        signature_policy['signed_by'] = proto_signature_policy.signed_by
    else:
        raise ValueError("Unknown signature policy type")
    return signature_policy


def decode_MSP_principal(proto_msp_principal):
    """Decodes MSP Principal

    :param proto_msp_principal: Bytes for MSP Principals
    :type proto_msp_principal: str
    :return: deserialized MSP Principal based on classification.
    """
    msp_principal = {}
    msp_principal['principal_classification'] = \
        proto_msp_principal.principal_classification
    if (msp_principal['principal_classification'] ==
            msp_principal_pb2.MSPPrincipal.ROLE):
        msp_principal['principal_classification'] = 'ROLE'
        proto_principal = msp_principal_pb2.MSPRole()
        proto_principal.ParseFromString(proto_msp_principal.principal)
        msp_principal['principal'] = {}
        msp_principal['principal']['msp_identifier'] = \
            proto_principal.msp_identifier
        if proto_principal.role == 0:
            msp_principal['principal']['role'] = 'MEMBER'
        elif proto_principal.role == 1:
            msp_principal['principal']['role'] = 'ADMIN'
        else:
            pass
    elif (msp_principal['principal_classification'] ==
          msp_principal_pb2.MSPPrincipal.ORGANIZATION_UNIT):
        msp_principal['principal_classification'] = 'ORGANIZATION_UNIT'
        proto_principal = msp_principal_pb2.OrganizationUnit()
        proto_principal.ParseFromString(proto_msp_principal.principal)
        msp_principal['principal'] = {}
        msp_principal['principal']['msp_identifier'] = \
            proto_principal.msp_identifier
        msp_principal['principal']['organizational_unit_identifier'] = \
            proto_principal.organizational_unit_identifier
        msp_principal['principal']['certifiers_identifier'] = \
            proto_principal.certifiers_identifier
    else:
        # Case of IDENTITY
        msp_principal = decode_identity(proto_msp_principal.principal)
    return msp_principal


def decode_config_signature(proto_configSignature):
    """Decodes Configuration Signature

    :param proto_configSignature: ConfigSignature() object
    :type proto_configSignature: str
    :return: deserialized config signature after header decode.
    """
    config_signature = {}
    config_signature['signature_header'] = \
        decode_signature_header(proto_configSignature.signature_header)
    config_signature['signature'] = proto_configSignature.signature
    return config_signature


def decode_fabric_MSP_config(msp_config_bytes):
    """Decodes Fabric MSP Configuration

    :param msp_config_bytes: Serialized configuration for MSP
    :type msp_config_bytes: str
    :return: Deserialized MSP configuration and certs.
    """
    msp_config = {}
    proto_msp_config = msp_config_pb2.FabricMSPConfig()
    proto_msp_config.ParseFromString(msp_config_bytes)
    msp_config['name'] = proto_msp_config.name
    msp_config['root_certs'] = to_PEM_certs(proto_msp_config.root_certs)
    msp_config['intermediate_certs'] = \
        to_PEM_certs(proto_msp_config.intermediate_certs)
    msp_config['admins'] = to_PEM_certs(proto_msp_config.admins)
    msp_config['revocation_list'] = \
        to_PEM_certs(proto_msp_config.revocation_list)
    msp_config['signing_identity'] = \
        decode_signing_identity_info(
            proto_msp_config.signing_identity.SerializeToString())
    msp_config['crypto_config'] = \
        decode_crypto_config(
            proto_msp_config.crypto_config.SerializeToString())
    ou_identifiers = [
        decode_fabric_OU_identifier(x) for x in
        proto_msp_config.organizational_unit_identifiers]
    msp_config['organizational_unit_identifiers'] = ou_identifiers
    msp_config['tls_root_certs'] = \
        to_PEM_certs(proto_msp_config.tls_root_certs)
    msp_config['tls_intermediate_certs'] = \
        to_PEM_certs(proto_msp_config.tls_intermediate_certs)
    msp_config['fabric_node_ous'] = \
        decode_fabric_Nodes_OUs(proto_msp_config.fabric_node_ous)
    return msp_config


def decode_fabric_OU_identifier(FabricOUIdentifier):
    """Decodes Fabric OU Identifier

    :param FabricOUIdentifier: OU Identifier
    :type FabricOUIdentifier: str
    :return: OU Identifier object.
    """

    return {
        'certificate': FabricOUIdentifier.certificate.decode(),
        'organizational_unit_identifier':
            FabricOUIdentifier.organizational_unit_identifier
    }


def decode_fabric_Nodes_OUs(proto_node_organizational_units):
    """Decodes Fabric Node OUs

    :param proto_node_organizational_units: OUs
    :type proto_node_organizational_units: str
    :return: deserialized list of OU Identifier objects.
    """
    node_organizational_units = {}

    if proto_node_organizational_units:
        node_organizational_units['enable'] = \
            proto_node_organizational_units.enable
        node_organizational_units['client_ou_identifier'] = \
            decode_fabric_OU_identifier(
                proto_node_organizational_units.client_ou_identifier)
        node_organizational_units['peer_ou_identifier'] = \
            decode_fabric_OU_identifier(
                proto_node_organizational_units.peer_ou_identifier)

    return node_organizational_units


def to_PEM_certs(buffer_array_in):
    """Decodes String buffer input to PEM Certs

    :param buffer_array_in: certificate contents buffer
    :type buffer_array_in: str
    :return: Concats buffer contents and returns certs
    """
    return [b64encode(x).decode() for x in buffer_array_in]


def decode_signing_identity_info(signing_identity_info_bytes):
    """Decodes Signing identity information from MSP Configuration

    :param signing_identity_info_bytes: Byte string of the identity info
    :type signing_identity_info_bytes: str
    :return: deserialized signing identity information.
    """
    if signing_identity_info_bytes == b'':
        return None

    signing_identity_info = {}

    if signing_identity_info_bytes is not None:
        proto_signing_identity_info = msp_config_pb2.SigningIdentityInfo()
        proto_signing_identity_info.ParseFromString(
            signing_identity_info_bytes)
        signing_identity_info['public_signer'] = \
            proto_signing_identity_info.public_signer.decode()
        signing_identity_info['private_signer'] = \
            decode_key_info(
                proto_signing_identity_info.private_signer.SerializeToString())
    return signing_identity_info


def decode_key_info(key_info_bytes):
    """Decodes Key Infor in MSP Configuration

    :param key_info_bytes: Byte information containing KeyInfo
    :type key_info_bytes: str
    :return: deserialized key information.
    """
    key_info = {}
    if key_info_bytes:
        proto_key_info = msp_config_pb2.KeyInfo()
        proto_key_info.ParseFromString(key_info_bytes)
        key_info['key_identifier'] = proto_key_info.key_identifier
        key_info['key_material'] = 'private'
    return key_info


def decode_crypto_config(crypto_config_bytes):
    """Decodes Crypto Config in MSP Configuration

    :param crypto_config_bytes: Byte information of FabricCyptoConfig
    :type crypto_config_bytes: str
    :return: deserialized key information.
    """
    crypto_config = {}
    if crypto_config_bytes:
        proto_crypto_config = msp_config_pb2.FabricCryptoConfig()
        proto_crypto_config.ParseFromString(crypto_config_bytes)
        crypto_config['signature_hash_family'] = proto_crypto_config. \
            signature_hash_family
        crypto_config['identity_identifier_hash_function'] = \
            proto_crypto_config.identity_identifier_hash_function
    return crypto_config


def decode_chaincode_action_payload(payload_bytes):
    """Decodes chaincode action payload from ChaincodeAction

    :param payload_bytes: Bytes buffer of the payload
    :type payload_bytes: str
    :return: deserialized payload information and action.
    """
    payload = {}
    proto_chaincode_action_payload = transaction_pb2.ChaincodeActionPayload()
    proto_chaincode_action_payload.ParseFromString(payload_bytes)
    payload['chaincode_proposal_payload'] = \
        decode_chaincode_proposal_payload(
            proto_chaincode_action_payload.chaincode_proposal_payload)
    payload['action'] = \
        decode_chaincode_endorsed_action(proto_chaincode_action_payload.action)
    return payload


def decode_chaincode_proposal_payload(chaincode_proposal_payload_bytes):
    """Decodes chaincode proposal payload from ChaincodeProposal

    :param chaincode_proposal_payload_bytes: Bytes of chaincode proposal
    :type chaincode_proposal_payload_bytes: str
    :return: deserialized chaincode proposal payload information
    """
    chaincode_proposal_payload = {}
    proto_chaincode_proposal_payload = proposal_pb2.ChaincodeProposalPayload()
    proto_chaincode_proposal_payload.ParseFromString(
        chaincode_proposal_payload_bytes)
    chaincode_proposal_payload['input'] = \
        proto_chaincode_proposal_payload.input
    # Transient map is not allowed to be included on the ledger.
    return chaincode_proposal_payload


def decode_chaincode_endorsed_action(proto_chaincode_endorsed_action):
    """Decodes chaincode endorsed action

    :param proto_chaincode_endorsed_action: Object containing endorsements
    :return: deserialized chaincode endorsement action.
    """
    action = {}
    action['proposal_response_payload'] = \
        decode_proposal_response_payload(
            proto_chaincode_endorsed_action.proposal_response_payload)
    action['endorsements'] = []
    for endorsement in proto_chaincode_endorsed_action.endorsements:
        endorsement = decode_endorsement(endorsement)
        action['endorsements'].append(endorsement)
    return action


def decode_endorsement(proto_endorsement):
    """Decodes each endorsement

    :param proto_endorsement: Object of endorsed content containing endorser
                           & related signature
    :return: deserialized endorsement content
    """
    endorsement = {}
    endorsement['endorser'] = decode_identity(proto_endorsement.endorser)
    endorsement['signature'] = proto_endorsement.signature
    return endorsement


def decode_proposal_response_payload(proposal_response_payload_bytes):
    """Decodes response payload in the proposal

    :param proposal_response_payload_bytes: Byte string of response payload
    :return: deserialized proposal response payload.
    """
    proposal_response_payload = {}
    proto_proposal_response_payload = \
        proposal_response_pb2.ProposalResponsePayload()
    proto_proposal_response_payload.ParseFromString(
        proposal_response_payload_bytes)
    proposal_response_payload['proposal_hash'] = \
        binascii.b2a_hex(proto_proposal_response_payload.proposal_hash)
    proposal_response_payload['extension'] = \
        decode_chaincode_action(proto_proposal_response_payload.extension)
    return proposal_response_payload


def decode_chaincode_action(action_bytes):
    """Decodes chaincode actions

    :param action_bytes: Byte buffer of the chaincode action
    :type action_bytes: str
    :return: deserialized chaincode action of results, events and response
    """
    chaincode_action = {}
    proto_chaincode_action = proposal_pb2.ChaincodeAction()
    proto_chaincode_action.ParseFromString(action_bytes)
    chaincode_action['results'] = \
        decode_readwrite_sets(proto_chaincode_action.results)
    chaincode_action['events'] = \
        decode_chaincode_events(proto_chaincode_action.events)
    chaincode_action['response'] = \
        decode_response(proto_chaincode_action.response)
    chaincode_action['chaincode_id'] = \
        decode_chaincode_id(proto_chaincode_action.chaincode_id)
    return chaincode_action


def decode_chaincode_events(event_bytes):
    """Decodes events in the chaincode

    :param event_bytes: Byte buffer of event content
    :type event_bytes: str
    :return: deserialized event contents.
    """
    events = {}
    proto_events = chaincode_event_pb2.ChaincodeEvent()
    proto_events.ParseFromString(event_bytes)
    events['chaincode_id'] = proto_events.chaincode_id
    events['tx_id'] = proto_events.tx_id
    events['event_name'] = proto_events.event_name
    events['payload'] = proto_events.payload
    return events


def decode_chaincode_id(proto_chaincode_id):
    """Decodes chaincode ID information

    :param proto_chaincode_id: Object containing chaincode details
    :return: deserialized chaincode ID with path, name and version.
    """
    chaincode_id = {}
    if not proto_chaincode_id:
        return chaincode_id
    chaincode_id['path'] = proto_chaincode_id.path
    chaincode_id['name'] = proto_chaincode_id.name
    chaincode_id['version'] = proto_chaincode_id.version
    return chaincode_id


def decode_readwrite_sets(rw_sets_bytes):
    """Decodes read write sets from a given TxReadWriteSet

    :param rw_sets_bytes: Byte buffer of the TxReadWriteSet
    :type rw_sets_bytes: str
    :return: deserialized transaction read write set contents.
    """
    proto_tx_read_write_set = rwset_pb2.TxReadWriteSet()
    proto_tx_read_write_set.ParseFromString(rw_sets_bytes)
    tx_read_write_set = {}
    tx_read_write_set['data_model'] = proto_tx_read_write_set.data_model
    if (tx_read_write_set['data_model'] == rwset_pb2.TxReadWriteSet.KV):
        tx_read_write_set['ns_rwset'] = []
        proto_ns_rwset = proto_tx_read_write_set.ns_rwset
        for rw_set in proto_ns_rwset:
            kv_rw_set = {}
            proto_kv_rw_set = rw_set
            kv_rw_set['namespace'] = proto_kv_rw_set.namespace
            kv_rw_set['rwset'] = decode_kv_rw_set(proto_kv_rw_set.rwset)
            tx_read_write_set['ns_rwset'].append(kv_rw_set)
    else:
        tx_read_write_set['ns_rwset'] = proto_tx_read_write_set.ns_rwset
    return tx_read_write_set


def decode_kv_rw_set(kv_bytes):
    """Decodes Key Value Read Write Set from KV Bytes

    :param kv_bytes: Buffer of key value bytes
    :type kv_bytes: str
    :return: deserialized key value read write set of reads, writes
             and range queries information.
    """
    proto_kv_rw_set = kv_rwset_pb2.KVRWSet()
    proto_kv_rw_set.ParseFromString(kv_bytes)
    kv_rw_set = {}
    # KV readwrite set has 3 arrays
    kv_rw_set['reads'] = []
    kv_rw_set['range_queries_info'] = []
    kv_rw_set['writes'] = []
    # Build reads
    reads = kv_rw_set['reads']
    proto_reads = proto_kv_rw_set.reads
    for read in proto_reads:
        reads.append(decode_kv_read(read))
    # Build range_queries_info
    range_queries_info = kv_rw_set['range_queries_info']
    proto_range_queries_info = proto_kv_rw_set.range_queries_info
    for range_query in proto_range_queries_info:
        range_queries_info.append(decode_range_query_info(range_query))
    # Build writes
    writes = kv_rw_set['writes']
    proto_writes = proto_kv_rw_set.writes
    for write in proto_writes:
        writes.append(decode_kv_write(write))

    kv_rw_set['reads'] = reads
    kv_rw_set['range_queries_info'] = range_queries_info
    kv_rw_set['writes'] = writes
    return kv_rw_set


def decode_kv_read(proto_kv_read):
    """Decodes Key Value Read

    :param proto_kv_read: Object of the key value with read contents
    :return: deserialized key value read contents with block num and tx_num
    """
    kv_read = {}
    kv_read['key'] = proto_kv_read.key
    proto_version = proto_kv_read.version
    if proto_version:
        kv_read['version'] = {}
        kv_read['version']['block_num'] = str(proto_version.block_num)
        kv_read['version']['tx_num'] = str(proto_version.tx_num)
    else:
        kv_read['version'] = None
    return kv_read


def decode_range_query_info(proto_range_query_info):
    """Decodes range query information from KV RW sets.

    :param proto_range_query_info: Object of key value read write range queries
    :return: deserialized range query information with merkle hashes.
    """
    range_query_info = {}
    range_query_info['start_key'] = proto_range_query_info.start_key
    range_query_info['end_key'] = proto_range_query_info.end_key
    range_query_info['itr_exhausted'] = proto_range_query_info.itr_exhausted

    proto_raw_reads = proto_range_query_info.raw_reads
    if proto_raw_reads:
        range_query_info['raw_reads'] = {}
        range_query_info['raw_reads']['kv_reads'] = []
        kv_read = None
        for kv_read in proto_raw_reads.kv_reads:
            kv_read = decode_kv_read(kv_read)
            range_query_info['raw_reads']['kv_reads'].append(kv_read)

    proto_reads_merkle_hashes = proto_range_query_info.reads_merkle_hashes
    if proto_reads_merkle_hashes:
        range_query_info['reads_merkle_hashes'] = {}
        range_query_info['reads_merkle_hashes']['max_degree'] = \
            proto_reads_merkle_hashes.max_degree
        range_query_info['reads_merkle_hashes']['max_level'] = \
            proto_reads_merkle_hashes.max_level
        range_query_info['reads_merkle_hashes']['max_level_hashes'] = \
            proto_reads_merkle_hashes.max_level_hashes
    return range_query_info


def decode_kv_write(proto_kv_write):
    """Decodes key value write instance

    :param proto_kv_write: Object containing key value writes
    :return: deserialized key value write contents and values.
    """
    kv_write = {}
    kv_write['key'] = proto_kv_write.key
    kv_write['is_delete'] = proto_kv_write.is_delete
    kv_write['value'] = proto_kv_write.value
    return kv_write


def decode_response(proto_response):
    """Decodes response containing status, message and payload

    :param proto_response: Object containing proto responses
    :return: deserialized response from protobuf objects
    """
    response = {}
    if proto_response:
        response['status'] = proto_response.status
        response['message'] = proto_response.message
        response['payload'] = proto_response.payload
    return response


def decode_fabric_peers_info(peers_info_bytes):
    """Decodes Fabric Peers Information

    :param peers_info_bytes: Serialized information about Peer
    :type peers_info_bytes: str
    :return: Deserialized Peers information and certs.
    """
    peers_info = []

    for peer_info_bytes in peers_info_bytes:
        peer = {}

        # identity
        peer_identity = identities_pb2.SerializedIdentity()
        peer_identity.ParseFromString(peer_info_bytes.identity)
        if hasattr(peer_identity, 'mspid'):
            peer['mspid'] = peer_identity.mspid
        if hasattr(peer_identity, 'id_bytes'):
            peer['id_bytes'] = peer_identity.id_bytes.decode()

        # state info
        peer_state_info = message_pb2.GossipMessage()
        peer_state_info.ParseFromString(peer_info_bytes.state_info.payload)

        if peer_state_info.state_info.properties:

            if hasattr(peer_state_info.state_info.properties, 'ledger_height'):
                peer['ledger_height'] = int(
                    peer_state_info.state_info.properties.ledger_height)

            if hasattr(peer_state_info.state_info.properties, 'chaincodes'):
                peer['chaincodes'] = []
                if peer_state_info.state_info.properties.chaincodes:
                    ccs = peer_state_info.state_info.properties.chaincodes
                    for chaincode in ccs:
                        cc = {}
                        cc['name'] = chaincode.name
                        cc['version'] = chaincode.version
                        peer['chaincodes'].append(cc)

        # membership info
        peer_membership_info = message_pb2.GossipMessage()
        membership_payload = peer_info_bytes.membership_info.payload
        peer_membership_info.ParseFromString(membership_payload)

        peer['endpoint'] = peer_membership_info.alive_msg.membership.endpoint

        peers_info.append(peer)

    return sorted(peers_info, key=lambda peer: peer['endpoint'])


def decode_fabric_endpoints(endpoints):
    """Decodes Fabric Endpoints

    :param endpoints: Fabric Endpoints
    :type endpoints: str
    :return: Deserialized endpoints.
    """

    endpoints_info = []
    for item in endpoints:
        endpoint = {}

        endpoint['host'] = item.host
        endpoint['port'] = int(item.port)

        endpoints_info.append(endpoint)
    return endpoints_info



# Copyright IBM Corp. 2017 All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest

from hfc.fabric.channel.channel_configuration import ChannelConfiguration

CONFIG_FILE_PATH = os.path.join(
    os.path.dirname(__file__),
    '../../fixtures/e2e_cli/channel-artifacts/channel.tx')


class ChannelConfigurationTest(unittest.TestCase):
    """Test for channel configuration. """

    def setUp(self):
        self.file_path = CONFIG_FILE_PATH

    def test_create_channel_configuration_by_file(self):
        chan_conf = ChannelConfiguration(file_path=self.file_path)
        self.assertIsNotNone(chan_conf.config)

    def test_create_channel_configuration_by_bytes(self):
        with open(self.file_path, mode="rb") as file:
            config = file.read()

        chan_conf = ChannelConfiguration(config=config)
        self.assertIsNotNone(chan_conf.config)

    def test_create_channel_configuration(self):
        chan_conf = ChannelConfiguration()
        self.assertIsNone(chan_conf.config)

    def test_set_channel_configuration(self):
        chan_conf = ChannelConfiguration()
        self.assertIsNone(chan_conf.config)

        with open(self.file_path, mode="rb") as file:
            config = file.read()
        chan_conf.config = config
        self.assertIsNotNone(chan_conf.config)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
#

# Set default logging handler to avoid "No handler found" warnings.
import logging
from .client import Client  # noqa

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# SPDX-License-Identifier: Apache-2.0

import logging

from hfc.fabric.channel.channel import Channel
from hfc.fabric_network.contract import Contract

consoleHandler = logging.StreamHandler()
_logger = logging.getLogger(__name__)

_logger.setLevel(logging.DEBUG)
_logger.addHandler(consoleHandler)


class Network(object):
    """A Network represents the set of peers in a Fabric network.
    Applications should get a Network instance using the
    gateway's getNetwork method.
    """

    def __init__(self, gateway, channel):
        """ Construct Network. """
        self.gateway = gateway
        self.channel = channel
        self.contracts = dict()
        self.initialized = False
        self.listeners = dict()
        self.discovery_enabled = False

    async def __init_internal_channel(self, discovery):
        """
        Initialize the channel if it hasn't been done
        :param discovery: must include requestor
        :return:
        """
        if discovery:
            self.discovery_enabled = True
            client = self.gateway.get_client()
            ledger_peers = client._peers
            if len(ledger_peers) == 0:
                _logger.error("No peers defined")
        else:
            if not isinstance(self.channel, Channel):
                _logger.error("network.channel should be an instance of channel")
            ledger_peers = self.channel._peers
            if len(ledger_peers) == 0:
                _logger.error("No peers defined")

        success = False

        for ledger_peer in ledger_peers:
            try:
                await self.gateway.client.init_with_discovery(discovery['requestor'],
                                                              ledger_peers[ledger_peer],
                                                              self.channel)
                success = True
                self.channel = self.gateway.client.get_channel(self.channel)
            except Exception:
                _logger.warning('Unable to initialize channel. Attempted to contact %s Peers. Last error was %s',
                                ledger_peer,
                                Exception)
            if success:
                break

    async def _initialize(self, discover=None):
        """
        Initialize this network instance
        :param discover:
        :return:
        """
        if self.initialized:
            return
        await self.__init_internal_channel(discover)
        self.initialized = True

    def get_contract(self, chaincode_id):
        if not self.initialized:
            _logger.error("Unable to get contract as network has failed to initialize")

        if chaincode_id not in self.contracts:
            contract = Contract(self, chaincode_id, self.gateway)
            self.contracts[chaincode_id] = contract

        return self.contracts[chaincode_id]



# Copyright 281165273@qq.com. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0
import logging
import random
import sys
import re
from _sha256 import sha256

from hfc.protos.msp import msp_principal_pb2

from hfc.fabric.block_decoder import BlockDecoder
from hfc.fabric.transaction.tx_proposal_request import \
    create_tx_prop_req, CC_INSTALL, CC_TYPE_GOLANG, \
    CC_INSTANTIATE, CC_UPGRADE, CC_INVOKE, CC_QUERY
from hfc.protos.common import common_pb2, policies_pb2, collection_pb2
from hfc.protos.orderer import ab_pb2
from hfc.protos.peer import chaincode_pb2, proposal_pb2
from hfc.protos.discovery import protocol_pb2
from hfc.protos.utils import create_cc_spec, create_seek_info, \
    create_seek_payload, create_envelope
from hfc.util import utils
from hfc.util.utils import proto_str, current_timestamp, proto_b, \
    build_header, build_channel_header, build_cc_proposal, \
    send_transaction_proposal, pem_to_der, package_chaincode
from .channel_eventhub import ChannelEventHub

SYSTEM_CHANNEL_NAME = "testchainid"

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


class Channel(object):
    """The class represents the channel.

    This is a client-side-only call. To create a new channel in the fabric
    call client._create_or_update_channel().
    """

    def __init__(self, name, client):
        """
        Construct channel instance

        :param client: fabric client instance, which provides operational context
        :param name: a unique name serves as the identifier of the channel
        :return: An instance of channel
        """
        pat = "^[a-z][a-z0-9.-]*$"  # matching patter for regex checker
        if not re.match(pat, name):
            raise ValueError(
                "ERROR: Channel name is invalid. It should be a \
                    string and match {}, but got {}".format(pat, name)
            )

        self._name = name
        self._client = client
        self._orderers = {}
        self._peers = {}
        # enable communication between peers of different orgs and discovery
        self._anchor_peers = []
        self._kafka_brokers = []
        # self._msp_manager = MSPManger() # TODO: use something instead
        self._initialized = False
        self._is_dev_mode = False
        self._channel_event_hubs = {}

    def add_orderer(self, orderer):
        """Add orderer endpoint to a channel object.

        A channel instance may choose to use a single orderer node, which
        will broadcast requests to the rest of the orderer network. Or if
        the application does not trust the orderer nodes, it can choose to
        use more than one by adding them to the channel instance. And all
        APIs concerning the orderer will broadcast to all _orderers
        simultaneously.

        :param orderer: an instance of the Orderer class
        :return:
        """
        self._orderers[orderer.endpoint] = orderer

    def remove_orderer(self, orderer):
        """Remove orderer endpoint from a channel object.

        :param orderer: an instance of the Orderer class
        :return:
        """
        if orderer.endpoint in self._orderers:
            self._orderers.pop(orderer.endpoint, None)

    def add_peer(self, peer):
        """Add peer endpoint to a chain object.

        :param peer: an instance of the Peer class
        :return:
        """
        self._peers[peer.endpoint] = peer

    def remove_peer(self, peer):
        """Remove peer endpoint from a channel object.

        :param peer: an instance of the Peer class
        :return:
        """
        if peer.endpoint in self._peers:
            self._peers.pop(peer.endpoint, None)

    @property
    def orderers(self):
        """Get _orderers of a channel.

        :return: The orderer list on the channel
        """
        return self._orderers

    @property
    def peers(self):
        """Get peers of a channel.

        :return: The peer list on the chain
        """
        return self._peers

    @property
    def is_dev_mode(self):
        """Get is_dev_mode

        :return: is_dev_mode
        """
        return self._is_dev_mode

    @is_dev_mode.setter
    def is_dev_mode(self, mode):
        self._is_dev_mode = mode

    def _get_latest_block(self, tx_context, orderer):
        """Get latest block from orderer.

        :param tx_context: a tx_context instance
        :param orderer: a orderer instance
        :return:
        """
        seek_info = ab_pb2.SeekInfo()
        seek_info.start.newest = ab_pb2.SeekNewest()
        seek_info.stop.newest = ab_pb2.SeekNewest()
        seek_info.behavior = \
            ab_pb2.SeekInfo.SeekBehavior.Value('BLOCK_UNTIL_READY')

        seek_info_header = self._build_channel_header(
            common_pb2.HeaderType.Value('DELIVER_SEEK_INFO'),
            tx_context.tx_id, self._name, current_timestamp(),
            tx_context.epoch)

        signature_header = common_pb2.SignatureHeader()
        signature_header.creator = tx_context.identity
        signature_header.nonce = tx_context.nonce

        seek_payload = common_pb2.Payload()
        seek_payload.header.signature_header = \
            signature_header.SerializeToString()
        seek_payload.header.channel_header = \
            seek_info_header.SerializeToString()
        seek_payload.data = seek_info.SerializeToString()

        envelope = common_pb2.Envelope()
        envelope.signature = tx_context.sign(seek_payload.SerializeToString())
        envelope.payload = seek_payload.SerializeToString()

    def _get_random_orderer(self):
        if sys.version_info < (3, 0):
            return random.choice(self._orderers.values())
        else:
            return random.choice(list(self._orderers.values()))

    @property
    def name(self):
        """Get channel name.

        :return: channel name
        """
        return self._name

    def state_store(self):
        """Get the key val store instance of the instantiating client.
        Get the KeyValueStore implementation (if any)
        that is currently associated with this channel
        :return: the current KeyValueStore associated with this
        channel / client.


        """
        return self._client.state_store

    def _validate_state(self):
        """Validate channel state.


        :raises ValueError:

        """
        if not self._initialized:
            raise ValueError(
                "Channel {} has not been initialized.".format(self._name))

    @property
    def is_sys_chan(self):
        """Get if system channel"""
        return self._is_sys_chan

    def _validate_peer(self, peer):
        """Validate peer

        :param peer: peer
        :raises ValueError:

        """
        if not peer:
            raise ValueError("Peer value is null.")

        if self._is_sys_chan:
            return

        if peer not in self._peers.values():
            raise ValueError(
                f"Channel {self._name} does not have peer {peer.endpoint}")

        if self not in peer.channels:
            raise ValueError(
                f"Peer {peer.endpoint} not joined this channel {self._name}")

    def _validate_peers(self, peers):
        """Validate peer set

        :param peers: peers
        :raises ValueError:

        """
        if not peers:
            raise ValueError("Collection of peers is null.")

        if len(peers) == 0:
            raise ValueError("Collection of peers is empty.")

        for peer in peers:
            self._validate_peer(peer)

    def send_install_proposal(self, tx_context, peers=None):
        """Send install chaincode proposal

        :param install_proposal_req: install proposal request
        :param targets: a set of peer to send
        :param tx_context: a tx_context instance
        :param peers: peers (Default value = None)
        :return: a set of proposal response
        """
        if peers is None:
            targets = self._peers.values()
        else:
            targets = peers
        # self._validate_state() # TODO: enable this later
        # self._validate_peers(targets)  # TODO: enable this later

        if not tx_context:
            raise ValueError("InstallProposalRequest is null.")

        cc_deployment_spec = chaincode_pb2.ChaincodeDeploymentSpec()
        cc_deployment_spec.chaincode_spec.type = \
            chaincode_pb2.ChaincodeSpec.Type.Value(
                utils.proto_str(tx_context.tx_prop_req.cc_type))
        cc_deployment_spec.chaincode_spec.chaincode_id.name = \
            proto_str(tx_context.tx_prop_req.cc_name)
        cc_deployment_spec.chaincode_spec.chaincode_id.path = \
            proto_str(tx_context.tx_prop_req.cc_path)
        cc_deployment_spec.chaincode_spec.chaincode_id.version = \
            proto_str(tx_context.tx_prop_req.cc_version)
        if not self._is_dev_mode:
            if not tx_context.tx_prop_req.packaged_cc:
                cc_deployment_spec.code_package = \
                    package_chaincode(
                        tx_context.tx_prop_req.cc_path,
                        tx_context.tx_prop_req.cc_type)
            else:
                cc_deployment_spec.code_package = \
                    tx_context.tx_prop_req.packaged_cc

        channel_header_extension = proposal_pb2.ChaincodeHeaderExtension()
        channel_header_extension.chaincode_id.name = \
            proto_str("lscc")
        channel_header = utils.build_channel_header(
            common_pb2.ENDORSER_TRANSACTION,
            tx_context.tx_id,
            '',
            utils.current_timestamp(),
            tx_context.epoch,
            channel_header_extension.SerializeToString()
        )

        header = utils.build_header(tx_context.identity,
                                    channel_header,
                                    tx_context.nonce)

        cci_spec = chaincode_pb2.ChaincodeInvocationSpec()
        cci_spec.chaincode_spec.type = \
            chaincode_pb2.ChaincodeSpec.Type.Value(CC_TYPE_GOLANG)
        cci_spec.chaincode_spec.chaincode_id.name = proto_str("lscc")
        cci_spec.chaincode_spec.input.args.extend(
            [proto_b(CC_INSTALL), cc_deployment_spec.SerializeToString()])
        proposal = utils.build_cc_proposal(
            cci_spec, header,
            tx_context.tx_prop_req.transient_map)
        signed_proposal = utils.sign_proposal(tx_context, proposal)

        responses = [peer.send_proposal(signed_proposal)
                     for peer in targets]

        return responses, proposal, header

    def _build_channel_header(type, tx_id, channel_id,
                              timestamp, epoch=0, extension=None):
        """Build channel.

        :param extension: extension (Default value = None)
        :param timestamp: timestamp
        :param channel_id: channel id
        :param tx_id: transaction id
        :param type: type
        :param epoch: epoch
        :return: common_proto.Header instance (Default value = 0)
        """
        channel_header = common_pb2.ChannelHeader()
        channel_header.type = type
        channel_header.version = 1
        channel_header.channel_id = proto_str(channel_id)
        channel_header.tx_id = proto_str(tx_id)
        channel_header.epoch = epoch
        channel_header.timestamp = timestamp
        if extension:
            channel_header.extension = extension

        return channel_header

    def is_readonly(self):
        """Check the channel if read-only

        Get the channel status to see if the underlying channel has been
        terminated, making it a read-only channel, where information
        (transactions and state_store) can be queried but no new transactions
        can be submitted.

        :return: True if the channel is read-only, False otherwise.
        """
        pass

    def join_channel(self, request):
        """To join the peer to a channel.

        :param request: request
        :return: A coroutine to handle thanks to asyncio with await asyncio.gather(*responses)
        """
        _logger.debug('channel_join - start')

        for key in ['targets', 'block', 'tx_context']:
            if key not in request:
                err_msg = "Missing parameter {}".format(key)
                _logger.error('channel_join error: {}'.format(err_msg))
                raise ValueError(err_msg)

        chaincode_input = chaincode_pb2.ChaincodeInput()
        chaincode_input.args.extend([proto_b("JoinChain"), request['block']])
        chaincode_id = chaincode_pb2.ChaincodeID()
        chaincode_id.name = proto_str("cscc")

        cc_spec = create_cc_spec(chaincode_input, chaincode_id, 'GOLANG')
        cc_invoke_spec = chaincode_pb2.ChaincodeInvocationSpec()
        cc_invoke_spec.chaincode_spec.CopyFrom(cc_spec)

        tx_context = request['tx_context']
        extension = proposal_pb2.ChaincodeHeaderExtension()
        extension.chaincode_id.name = proto_str('cscc')
        channel_header = build_channel_header(
            common_pb2.HeaderType.Value('ENDORSER_TRANSACTION'),
            tx_context.tx_id,
            '',
            current_timestamp(),
            tx_context.epoch,
            extension=extension.SerializeToString())

        header = build_header(tx_context.identity,
                              channel_header,
                              tx_context.nonce)
        proposal = build_cc_proposal(cc_invoke_spec,
                                     header,
                                     request['transient_map'])

        return send_transaction_proposal(proposal,
                                         tx_context,
                                         request['targets'])

    def send_instantiate_proposal(self, tx_context, peers):
        """Send instantiate chaincode proposal.

        :param tx_context: transaction context
        :param peers: peers to send this proposal
        :return: True in success False in failure
        """
        if not peers:
            peers = self.peers.values()
        if not tx_context:
            raise Exception("The transaction context is null.")

        return self._send_cc_proposal(tx_context, CC_INSTANTIATE, peers)

    def send_upgrade_proposal(self, tx_context, peers):
        """Upgrade the chaincode.

        :param tx_context: transaction context
        :param peers: peers to send this proposal
        :return: True in success and False in failure
        """

        if not peers:
            peers = self.peers.values()
        if not tx_context:
            raise Exception("The transaction context is null.")

        return self._send_cc_proposal(tx_context, CC_UPGRADE, peers)

    def _build_principal(self, identity):
        if 'role' not in identity:
            raise Exception('NOT IMPLEMENTED')

        newPrincipal = msp_principal_pb2.MSPPrincipal()

        newPrincipal.principal_classification = \
            msp_principal_pb2.MSPPrincipal.ROLE

        newRole = msp_principal_pb2.MSPRole()

        roleName = identity['role']['name']
        if roleName == 'peer':
            newRole.role = msp_principal_pb2.MSPRole.PEER
        elif roleName == 'member':
            newRole.role = msp_principal_pb2.MSPRole.MEMBER
        elif roleName == 'admin':
            newRole.role = msp_principal_pb2.MSPRole.ADMIN
        else:
            raise Exception(f'Invalid role name found: must'
                            f' be one of "peer", "member" or'
                            f' "admin", but found "{roleName}"')

        mspid = identity['role']['mspId']
        if not mspid or not isinstance(mspid, str):
            raise Exception(f'Invalid mspid found: "{mspid}"')
        newRole.msp_identifier = mspid.encode()

        newPrincipal.principal = newRole.SerializeToString()

        return newPrincipal

    def _get_policy(self, policy):
        type = list(policy.keys())[0]
        # signed-by case
        if type == 'signed-by':
            signedBy = policies_pb2.SignaturePolicy()
            signedBy.signed_by = policy['signed-by']
            return signedBy
        # n-of case
        else:
            n = int(type.split('-of')[0])

            nOutOf = policies_pb2.SignaturePolicy.NOutOf()
            nOutOf.n = n
            subs = []
            for sub in policy[type]:
                subPolicy = self._get_policy(sub)
                subs.append(subPolicy)

            nOutOf.rules.extend(subs)

            nOf = policies_pb2.SignaturePolicy()
            nOf.n_out_of.CopyFrom(nOutOf)

            return nOf

    def _check_policy(self, policy):
        if not policy:
            raise Exception('Missing Required Param "policy"')

        if 'identities' not in policy \
                or policy['identities'] == '' \
                or not len(policy['identities']):
            raise Exception('Invalid policy, missing'
                            ' the "identities" property')
        elif not isinstance(policy['identities'], list):
            raise Exception('Invalid policy, the "identities"'
                            ' property must be an array')

        if 'policy' not in policy \
                or policy['policy'] == '' \
                or not len(policy['policy']):
            raise Exception('Invalid policy, missing the'
                            ' "policy" property')

    def _build_policy(self, policy, msps=None, returnProto=False):
        proto_signature_policy_envelope = \
            policies_pb2.SignaturePolicyEnvelope()

        if policy:
            self._check_policy(policy)
            proto_signature_policy_envelope.version = 0
            proto_signature_policy_envelope.rule.CopyFrom(
                self._get_policy(policy['policy']))
            proto_signature_policy_envelope.identities.extend(
                [self._build_principal(x) for x in policy['identities']])
        else:
            # TODO need to support MSPManager
            # no policy was passed in, construct a 'Signed By any member
            # of an organization by mspid' policy
            # construct a list of msp principals to select from using the
            # 'n out of' operator

            # for not making it fail with current code
            return proto_b('')

            principals = []
            signedBys = []
            index = 0

            if msps is None:
                msps = []

            for msp in msps:
                onePrn = msp_principal_pb2.MSPPrincipal()
                onePrn.principal_classification = \
                    msp_principal_pb2.MSPPrincipal.ROLE

                memberRole = msp_principal_pb2.MSPRole()
                memberRole.role = msp_principal_pb2.MSPRole.MEMBER
                memberRole.msp_identifier = msp

                onePrn.principal = memberRole.SerializeToString()

                principals.append(onePrn)

                signedBy = policies_pb2.SignaturePolicy()
                index += 1
                signedBy.signed_by = index
                signedBys.append(signedBy)

            if len(principals) == 0:
                raise Exception('Verifying MSPs not found in the'
                                ' channel object, make sure'
                                ' "initialize()" is called first.')

            oneOfAny = policies_pb2.SignaturePolicy.NOutOf()
            oneOfAny.n = 1
            oneOfAny.rules.extend(signedBys)

            noutof = policies_pb2.SignaturePolicy()
            noutof.n_out_of.CopyFrom(oneOfAny)

            proto_signature_policy_envelope.version = 0
            proto_signature_policy_envelope.rule.CopyFrom(noutof)
            proto_signature_policy_envelope.identities.extend(principals)

        if returnProto:
            return proto_signature_policy_envelope

        return proto_signature_policy_envelope.SerializeToString()

    def _send_cc_proposal(self, tx_context, command, peers):

        args = []
        request = tx_context.tx_prop_req

        args.append(proto_b(request.fcn))
        for arg in request.args:
            args.append(proto_b(arg))

        # construct the deployment spec
        cc_id = chaincode_pb2.ChaincodeID()
        cc_id.name = request.cc_name
        cc_id.version = request.cc_version

        cc_input = chaincode_pb2.ChaincodeInput()
        cc_input.args.extend(args)
        cc_spec = create_cc_spec(cc_input, cc_id, CC_TYPE_GOLANG)

        cc_dep_spec = chaincode_pb2.ChaincodeDeploymentSpec()
        cc_dep_spec.chaincode_spec.CopyFrom(cc_spec)

        # Pass msps, TODO create an MSPManager as done in fabric-sdk-node
        policy = self._build_policy(request.cc_endorsement_policy)

        args = [
            proto_b(command),
            proto_b(self.name),
            cc_dep_spec.SerializeToString(),
            policy,
            proto_b('escc'),
            proto_b('vscc'),
        ]

        # collections_configs need V1_2 or later capability enabled,
        # otherwise private channel collections and data are not available
        collections_configs = []
        if request.collections_config:
            for config in request.collections_config:
                static_config = collection_pb2.StaticCollectionConfig()
                static_config.name = config['name']
                static_config.member_orgs_policy.signature_policy. \
                    CopyFrom(self._build_policy(config['policy'],
                                                returnProto=True))
                static_config.maximum_peer_count = config['maxPeerCount']
                static_config. \
                    required_peer_count = config.get('requiredPeerCount', 0)
                static_config.block_to_live = config.get('blockToLive', 0)
                static_config.member_only_read = config.get('memberOnlyRead',
                                                            False)

                collections_config = collection_pb2.CollectionConfig()
                collections_config.static_collection_config.CopyFrom(
                    static_config
                )

                collections_configs.append(collections_config)

            cc_coll_cfg = collection_pb2.CollectionConfigPackage()
            cc_coll_cfg.config.extend(collections_configs)
            args.append(cc_coll_cfg.SerializeToString())

        # construct the invoke spec
        invoke_input = chaincode_pb2.ChaincodeInput()
        invoke_input.args.extend(args)

        invoke_cc_id = chaincode_pb2.ChaincodeID()
        invoke_cc_id.name = proto_str('lscc')

        cc_invoke_spec = chaincode_pb2.ChaincodeInvocationSpec()
        cc_invoke_spec.chaincode_spec.CopyFrom(create_cc_spec(invoke_input,
                                                              invoke_cc_id,
                                                              CC_TYPE_GOLANG)
                                               )

        extension = proposal_pb2.ChaincodeHeaderExtension()
        extension.chaincode_id.name = proto_str('lscc')
        channel_header = build_channel_header(
            common_pb2.ENDORSER_TRANSACTION,
            tx_context.tx_id,
            self.name,
            current_timestamp(),
            epoch=0,
            extension=extension.SerializeToString()
        )

        header = build_header(tx_context.identity,
                              channel_header,
                              tx_context.nonce)
        proposal = build_cc_proposal(
            cc_invoke_spec,
            header,
            request.transient_map)

        signed_proposal = utils.sign_proposal(tx_context, proposal)
        response = [peer.send_proposal(signed_proposal)
                    for peer in peers]
        return response, proposal, header

    def send_tx_proposal(self, tx_context, peers):
        """Invoke the chaincode

        Send a transaction proposal to one or more endorser without
        creating a channel.
        :param tx_context: transaction context
        :param peers: the pees to send this proposal
                 if it is None the channel peers list will be used.
        channel_id: channel id
        client: client context
        :return: True in success or False in failure.
        """
        if not peers:
            peers = self.peers.values()

        return Channel._send_tx_proposal(self.name, tx_context, peers)

    @staticmethod
    def _send_tx_proposal(channel_id, tx_context, peers):

        request = tx_context.tx_prop_req

        args = []
        if request.fcn:
            args.append(proto_b(request.fcn))
        else:
            args.append(proto_b(CC_INVOKE))

        for arg in request.args:
            if isinstance(arg, bytes):
                args.append(arg)
            else:
                args.append(proto_b(arg))

        cc_id = chaincode_pb2.ChaincodeID()
        cc_id.name = request.cc_name
        if request.prop_type not in (CC_QUERY, CC_INVOKE):
            cc_id.version = request.cc_version

        cc_input = chaincode_pb2.ChaincodeInput()
        cc_input.args.extend(args)

        cc_spec = chaincode_pb2.ChaincodeSpec()
        cc_spec.type = chaincode_pb2.ChaincodeSpec.Type.Value(CC_TYPE_GOLANG)
        cc_spec.chaincode_id.CopyFrom(cc_id)
        cc_spec.input.CopyFrom(cc_input)

        extension = proposal_pb2.ChaincodeHeaderExtension()
        extension.chaincode_id.name = request.cc_name
        cc_invoke_spec = chaincode_pb2.ChaincodeInvocationSpec()
        cc_invoke_spec.chaincode_spec.CopyFrom(cc_spec)

        channel_header = build_channel_header(
            common_pb2.ENDORSER_TRANSACTION,
            tx_context.tx_id,
            channel_id,
            current_timestamp(),
            tx_context.epoch,
            extension=extension.SerializeToString())

        header = build_header(tx_context.identity,
                              channel_header,
                              tx_context.nonce)

        # chaincode real proposal
        proposal = build_cc_proposal(cc_invoke_spec, header,
                                     request.transient_map)
        signed_proposal = utils.sign_proposal(tx_context, proposal)
        responses = [peer.send_proposal(signed_proposal)
                     for peer in peers]

        # chaincode proposal without transient map
        # https://jira.hyperledger.org/browse/FAB-12536?focusedCommentId=52438&page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel#comment-52438 # noqa
        proposal = build_cc_proposal(cc_invoke_spec, header, None)

        return responses, proposal, header

    def query_instantiated_chaincodes(self, tx_context, peers,
                                      transient_map=None):
        """

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param transient_map: transient map
        :return: chain code response (Default value = None)
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='getchaincodes',
            cc_name='lscc',
            cc_type=CC_TYPE_GOLANG,
            args=[],
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def query_transaction(self, tx_context, peers, tx_id,
                          transient_map=None):
        """Queries the ledger for Transaction by transaction ID.

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param tx_id: transaction ID (string)
        :param transient_map: transient map
        :return: chain code response (Default value = None)
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetTransactionByID',
            cc_name='qscc',
            args=[self.name, tx_id],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def get_block_between(self, tx_context, orderer, start, end):
        """

        :param tx_context: tx_context instance
        :param orderer: orderer instance
        :param start: id of block to start query for
        :param end: id of block to end query for
        :return: block(s)
        """
        seek_info = create_seek_info(start, end)
        seek_info_header = build_channel_header(
            common_pb2.HeaderType.Value('DELIVER_SEEK_INFO'),
            tx_context.tx_id,
            self._name,
            current_timestamp(),
            tx_context.epoch)

        seek_header = build_header(
            tx_context.identity,
            seek_info_header,
            tx_context.nonce)

        seek_payload_bytes = create_seek_payload(seek_header, seek_info)
        sig = tx_context.sign(seek_payload_bytes)

        envelope = create_envelope(sig, seek_payload_bytes)
        response = orderer.delivery(envelope)

        if response[0].block is None or response[0].block == '':
            _logger.error("fail to get block start from %s to %s" %
                          (str(start), str(end)))
            return None

        _logger.info("get block successfully, start from %s to %s" %
                     (str(start), str(end)))

        return response[0].block

    def query_block(self, tx_context, peers, block_number,
                    transient_map=None):
        """Queries the ledger for Block by block number.

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param block_number: block to query for
        :param transient_map: transient map (Default value = None)
        :return: class BlockDecoder
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetBlockByNumber',
            cc_name='qscc',
            args=[self.name, block_number],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def query_block_by_hash(self, tx_context, peers, block_hash,
                            transient_map=None):
        """

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param block_hash: block to query for
        :param transient_map: transient map (Default value = None)
        :return: class ChaincodeQueryResponse
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetBlockByHash',
            cc_name='qscc',
            args=[self.name, block_hash],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def query_block_by_txid(self, tx_context, peers, tx_id,
                            transient_map=None):
        """

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param tx_id: transaction id
        :param transient_map: transient map (Default value = None)
        :return: class ChaincodeQueryResponse
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetBlockByTxID',
            cc_name='qscc',
            args=[self.name, tx_id],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def query_info(self, tx_context, peers, transient_map=None):
        """Query the information of channel

        Queries for various useful information on the state of the channel
        (height, known peers).

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param transient_map:  (Default value = None)
        :return: class ChaincodeQueryResponse channelinfo with height, currently the only useful information.
        """
        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetChainInfo',
            cc_name='qscc',
            args=[self.name],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    def get_channel_config(self, tx_context, peers,
                           transient_map=None):
        """Query the current config block for this channel

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :param transient_map:  (Default value = None)
        :return: class ChaincodeQueryResponse channelinfo with height, currently the only useful information.
        """

        request = create_tx_prop_req(
            prop_type=CC_QUERY,
            fcn='GetConfigBlock',
            cc_name='cscc',
            args=[self.name],
            cc_type=CC_TYPE_GOLANG,
            transient_map=transient_map)

        tx_context.tx_prop_req = request
        return self.send_tx_proposal(tx_context, peers)

    async def get_channel_config_with_orderer(self, tx_context, orderer, decode=True):
        """Query the current config block for this channel

        :param tx_context: tx_context instance
        :param peers: peers in the channel
        :return:class ChaincodeQueryResponse channelinfo with height, currently the only useful information.
        """

        seek_info = create_seek_info()

        kwargs = {}
        if orderer._client_cert_path:
            with open(orderer._client_cert_path, 'rb') as f:
                b64der = pem_to_der(f.read())
                kwargs['tls_cert_hash'] = sha256(b64der).digest()

        seek_info_header = build_channel_header(
            common_pb2.HeaderType.Value('DELIVER_SEEK_INFO'),
            tx_context.tx_id,
            self.name,
            current_timestamp(),
            tx_context.epoch,
            **kwargs
        )

        seek_header = build_header(
            tx_context.identity,
            seek_info_header,
            tx_context.nonce)

        seek_payload_bytes = create_seek_payload(seek_header, seek_info)
        sig = tx_context.sign(seek_payload_bytes)
        envelope = create_envelope(sig, seek_payload_bytes)

        # this is a stream response
        block = None
        stream = orderer.delivery(envelope)
        async for v in stream:
            if v.block is None or v.block == '':
                msg = "fail to get block"
                _logger.error(msg)
                raise Exception(msg)
            block = v.block
            break

        decoded_block = BlockDecoder().decode(block.SerializeToString())
        last_config = decoded_block['metadata']['metadata'][common_pb2.LAST_CONFIG]

        # if nor first block
        if 'index' in last_config['value']:
            seek_info = create_seek_info(last_config['value']['index'],
                                         last_config['value']['index'])
            seek_payload_bytes = create_seek_payload(seek_header, seek_info)
            sig = tx_context.sign(seek_payload_bytes)
            envelope = create_envelope(sig, seek_payload_bytes)

            block = None
            stream = orderer.delivery(envelope)
            async for v in stream:
                if v.block is None or v.block == '':
                    msg = "fail to get block"
                    _logger.error(msg)
                    raise Exception(msg)
                block = v.block
                break

            if not decode:
                return block.SerializeToString()
            decoded_block = BlockDecoder().decode(block.SerializeToString())

        if not decode:
            return block.SerializeToString()

        envelope = decoded_block['data']['data'][0]
        payload = envelope['payload']
        channel_header = payload['header']['channel_header']

        if channel_header['type'] != common_pb2.CONFIG:
            raise Exception(f'Block must be of type "CONFIG"'
                            f' ({common_pb2.CONFIG}), but got'
                            f' "{channel_header["type"]}" instead')

        config_envelope = payload['data']
        return config_envelope

    def _discovery(self, requestor, target,
                   local=False, config=False, interests=None):
        """Send a request from a target peer to discover information about the
         network

        :param requestor: a user to make the request
        :type requestor: instance
        :param target: target peer to send discovery request
        :type target: instance
        :param local: include local endpoints in the query (Default value = False)
        :type local: bool
        :param config: include channel configuration in the query (Default value = False)
        :type config: bool
        :param interests: interests about an endorsement for cc (Default value = None)
        :type interests: list
        :return: Response from Discovery Service
        """

        auth = protocol_pb2.AuthInfo()
        sig = utils.create_serialized_identity(requestor)
        auth.client_identity = sig
        # TODO: add tls certificate in client and there
        discovery_req = protocol_pb2.Request()
        discovery_req.authentication.CopyFrom(auth)
        queries = []

        if local:
            q = protocol_pb2.Query()
            queries.append(q)
            local_peers = protocol_pb2.LocalPeerQuery()
            q.local_peers.CopyFrom(local_peers)
            _logger.info("DISCOVERY: adding local peers query")
        else:
            # It gives us state info about the channel
            # in addition of LocalPeerQuery information
            q = protocol_pb2.Query()
            queries.append(q)
            q.channel = self._name
            peer_query = protocol_pb2.PeerMembershipQuery()
            q.peer_query.CopyFrom(peer_query)
            _logger.info("DISCOVERY: adding channel peers query")

        if config:
            q = protocol_pb2.Query()
            queries.append(q)
            q.channel = self._name

            config_query = protocol_pb2.ConfigQuery()
            q.config_query.CopyFrom(config_query)
            _logger.info("DISCOVERY: adding config query")

        if interests and len(interests) > 0:
            q = protocol_pb2.Query()
            queries.append(q)
            q.channel = self._name

            cc_interests = []
            for interest in interests:
                proto_interest = self._build_proto_cc_interest(interest)
                cc_interests.append(proto_interest)

            cc_query = protocol_pb2.ChaincodeQuery()
            cc_query.interests.extend(cc_interests)
            q.cc_query.CopyFrom(cc_query)
            _logger.info("DISCOVERY: adding chaincodes/collection query")

        discovery_req.queries.extend(queries)

        request_bytes = discovery_req.SerializeToString()
        sig = requestor.cryptoSuite.sign(requestor.enrollment.private_key,
                                         request_bytes)
        envelope = create_envelope(sig, request_bytes)

        return target.send_discovery(envelope)

    def _build_proto_cc_interest(self, interest):
        """Use a list of DiscoveryChaincodeCall to build an interest.

        :param interest:
        :return:
        """
        cc_calls = []
        try:
            for cc in interest['chaincodes']:
                cc_call = protocol_pb2.ChaincodeCall()

                if cc.get('name'):
                    if not isinstance(cc['name'], str):
                        raise ValueError("chaincode names must be a string")
                    cc_call.name = cc['name']

                if cc.get('collection_names'):
                    if not isinstance(cc['collection_names'], list):
                        raise ValueError(
                            "collection_names must be an array of strings")
                    if not all(isinstance(x, str)
                               for x in cc['collection_names']):
                        raise ValueError("collection name must be a string")
                    cc_call.collection_names.extend(cc['collection_names'])

                cc_calls.append(cc_call)

        except AttributeError as e:
            _logger.error("The key 'chaincodes' is missing, {}".format(e))
            raise

        except KeyError as e:
            _logger.error("The key is missing, {}".format(e))
            raise

        interest_proto = protocol_pb2.ChaincodeInterest()
        interest_proto.chaincodes.extend(cc_calls)

        return interest_proto

    def newChannelEventHub(self, peer, requestor):
        channel_event_hub = ChannelEventHub(peer, self._name, requestor)
        if requestor.org not in self._channel_event_hubs:
            self._channel_event_hubs[requestor.org] = [channel_event_hub]
        else:
            self._channel_event_hubs[requestor.org].append(channel_event_hub)
        return channel_event_hub

    def getChannelEventHubsForOrg(self, requestor, mspid=None):
        if mspid:
            che = self._channel_event_hubs.get(mspid, [])
        else:
            che = self._channel_event_hubs.get(requestor.msp_id, [])

        return [x for x in che if x.connected]


def create_system_channel(client, name=SYSTEM_CHANNEL_NAME):
    """Create system channel instance

    :param client: client instance
    :param name: system channel name (Default value = SYSTEM_CHANNEL_NAME)
    :return: return system channel instance
    """
    return Channel(name, client, True)


def create_app_channel(client, name):
    """Create application channel instance

    :param client: client instance
    :param name: return application channel instance
    :return: system channel instance
    """
    return Channel(name, client, False)
